import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter_hooks/flutter_hooks.dart';
import 'package:flutter_query/flutter_query.dart';
import 'package:flutter_query/src/query.dart';
import 'package:flutter_query/src/query_cache.dart';
import 'package:flutter_query/src/query_observer.dart';
import 'package:provider/provider.dart';

typedef Future<T> QueryFunction<T>();
typedef Future PageableQueryFunction<T>(int page);
typedef dynamic GetFetchMoreFunction(latestPage, allPages);

typedef void QueryCallback<T>(T data);

QueryResult<T> useQuery<T>(
  String key,
  QueryFunction<T> fn, {
  QueryCallback onSuccess,
  QueryCallback onError,
  QueryCallback onSettled,
}) {
  return use(
    _QueryHook<T>(
      QueryConfig(
        queryKey: key,
        fn: fn,
        onSuccess: onSuccess,
        onError: onError,
        onSettled: onSettled,
      ),
    ),
  );
}

QueryResult<List> useInfiniteQuery<T>(
  String key,
  PageableQueryFunction<T> fn, {
  GetFetchMoreFunction getFetchMore,
  QueryCallback onSuccess,
  QueryCallback onError,
  QueryCallback onSettled,
}) {
  return use(
    _QueryHook<List>(
      QueryConfig(
        queryKey: key,
        fn: fn,
        onSuccess: onSuccess,
        onError: onError,
        onSettled: onSettled,
        getFetchMore: getFetchMore,
        isInfinite: true,
      ),
    ),
  );
}

class _QueryHook<T> extends Hook<QueryResult<T>> {
  const _QueryHook(this.config);

  final QueryConfig config;

  @override
  _QueryStateHook<T> createState() => _QueryStateHook();
}

class _QueryStateHook<T> extends HookState<QueryResult<T>, _QueryHook<T>> {
  QueryResult<T> _result;
  Function _unsubscribe;

  @override
  void initHook() {
    super.initHook();
    final cache = Provider.of<QueryCache>(context, listen: false);
    _unsubscribe = cache.watchQuery(
      hook.config,
      (QueryResult<T> result) {
        if (_result == null) {
          setState(() {
            _result = result;
          });
        } else {
          // TODO: IMPORTANT!! This is hacky! It is here because calling setState normally sometimes wouldn't cause a rebuild (while running the tests)
          // It is probably due to weird behavior or setup in the stream subscription in the queue observer. It could have to do with the sync flag?
          // Either way this is something that MUST be addressed and looked into. That is why I am making such a big important comment
          Future(() {
            setState(() {
              _result = result;
            });
          });
        }
      },
    );
  }

  @override
  void dispose() {
    super.dispose();
    _unsubscribe();
  }

  @override
  QueryResult<T> build(BuildContext context) {
    return _result;
  }
}
